<?php
# $Id: database.inc,v 1.17 2003/03/05 05:51:52 james Exp james $
#
#   database.inc, prepaid accounting, database access functions
#   Copyright (C) 2001  James Cameron (quozl@us.netrek.org)
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   Filesystem data storage
#
#	/var/ppa
#	/var/ppa/lock - lock file, always exists
#	/var/ppa/new - temporary file
#	/var/ppa/accounts - directory containing serialized arrays
#	/var/ppa/wtmps - directory containing processed wtmp files
#
#   Filesystem installation steps
#
#	mkdir /var/ppa
#       mkdir /var/ppa/accounts
#       mkdir /var/ppa/wtmps
#       touch /var/ppa/lock
#       chown -R www-data:www-data /var/ppa
#
#   Account array layout
#
#	$account['name']
#		account name
#	$account['address']
#		account address
#	$account['rate']
#		rate per hour of usage in dollars
#	$account['transactions'][0]['when']
#		timestamp of transaction creation
#		for usage transactions, the logout event time
#	$account['transactions'][0]['code']
#		item code of transaction 
#		for summarising
#	$account['transactions'][0]['comment']
#		narrative details of transaction
#		for usage transactions, includes login event time
#	$account['transactions'][0]['amount']
#		amount of transaction
#

if (is_file('/var/ppa/ppa.conf')) include('/var/ppa/ppa.conf');

# title graphic for each page
if (!defined('TITLE'))
    define('TITLE', '<img src=gic-title.jpg width=517 height=31 alt="Gilgandra Community Technology Centre">');

# directory in which data is to be put, see database.inc
if (!defined('PATH_ACCOUNTS'))
    define('PATH_ACCOUNTS', '/var/ppa/accounts/');

# file for locking access to database
if (!defined('PATH_LOCK'))
    define('PATH_LOCK', '/var/ppa/lock');

# file to use for new record before renaming to PATH_ACCOUNTS
if (!defined('PATH_NEW'))
    define('PATH_NEW', '/var/ppa/new');

# file to store item codes in
if (!defined('PATH_CODES'))
    define('PATH_CODES', '/var/ppa/codes');

# colours
if (!isset($colour_heading_one))       $colour_heading_one       = '#d0d0ef';
if (!isset($colour_heading_two))       $colour_heading_two       = '#e0e0ff';
if (!isset($colour_detail_debit_one))  $colour_detail_debit_one  = '#ffe0e0';
if (!isset($colour_detail_debit_two))  $colour_detail_debit_two  = '#efd0d0';
if (!isset($colour_detail_credit_one)) $colour_detail_credit_one = '#e0ffe0';
if (!isset($colour_detail_credit_two)) $colour_detail_credit_two = '#d0efd0';


# amount functions

# translate to integer scaled amount from string
function amount_from_string ($string) {
    return round(((float) $string) * 100.0);
}

# translate to string from integer scaled amount 
function amount_to_string ($amount) {
    return sprintf ("%01.2f", (((float)$amount) / 100.0));
}

# round to the nearest 5c mark, $2.23 goes to $2.25
function amount_round_5c ($amount) {
    return round(round($amount / 5.0) * 5.0);
}

# date functions

# return the default date
function date_default () {
    return time();
}

# translate a unix date to human readable ISO8601 format
function date_to_string ($date) {
    return date('Y-m-d H:i:s', $date);
}

# translate a human readable ISO8601 date to unix date
function date_from_string ($string) {
    list($date, $time) = explode(' ', $string);
    list($year, $month, $day) = explode('-', $date);
    list($hour, $minute, $second) = explode(':', $time);
    $date = mktime($hour, $minute, $second, $month, $day, $year);
    if (date_to_string($date) == $string) return $date;
    return 0;
}


# account functions

# translate account name to file path
function account_to_file ($name) {
    return (PATH_ACCOUNTS.$name);
}

# check if an account exists, returns true or false
function account_exists ($name) {
    return (is_file(account_to_file($name)));
}

# return a list of accounts known in the database
function account_list () {

    $directory = PATH_ACCOUNTS;
    $handle = opendir($directory);
    while (true) {
        $file = readdir($handle);
	if (gettype($file) == 'boolean') break;
	if ($file == '') break;
	if ($file == '.') continue;
	if ($file == '..') continue;
        $array[] = basename($file);
    }
    closedir($handle);

    if (gettype($array) == 'array') return $array;

    return array();
}

# write an account from array to file, with interlock
function account_write ($account) {

    # obtain interlock
    $lock = fopen(PATH_LOCK, 'a+');
    if (!flock($lock, 2)) {
	echo 'Failed to lock database';
	fclose($lock);
	return 0;
    }

    # open the file
    if (!($fp = fopen(PATH_NEW, 'w'))) {
	echo 'Failed to open database';
	fclose($lock);
	return 0;
    }

    # write the array to the file
    if (!fwrite($fp,serialize($account))) {
	echo 'Failed to write to database';
	fclose($fp);
	fclose($lock);
	return 0;
    }

    # close the file 
    if (!fclose($fp)) {
	echo 'Failed to close database';
	fclose($lock);
	return 0;
    }

    # derive the file name
    $file = account_to_file($account['name']);

    # rename the new file into place
    if (!rename(PATH_NEW, $file)) {
	echo 'Failed to move database file into place';
	fclose($lock);
	return 0;
    }

    # release the lock and return success
    fclose($lock);
    return 1;
}

# create an account
function account_create ($name, $address, $rate) {
    $account['name'] = strtolower($name);
    $account['address'] = $address;
    $account['rate'] = $rate;
    return account_write ($account);
}

# read an account from file into array
function account_read ($name) {

    # derive the file name
    $file = account_to_file($name);

    # return zero if no file
    if (!is_file($file)) { return 0; }

    # open the file
    if (!($fp = fopen($file, 'r'))) {
	echo 'Failed to open database';
	return 0;
    }

    # read the file
    $account = unserialize(fread($fp, filesize($file)));

    # close the file
    fclose($fp);

    return $account;
}

# return a default account
function account_default($name) {
    $account['name'] = $name;
    $account['address'] = "";
    $account['rate'] = amount_from_string("1.50");
    return $account;
}

# check for a duplicate transaction in an account
function account_find_transaction($account, $transaction) {
    if (isset($account['transactions'])) {
	reset ($account['transactions']);
	while (list ($key, $val) = each ($account['transactions'])) {
	    if ($account['transactions'][$key]['when'] != 
		$transaction['when']) continue;
	    if ($account['transactions'][$key]['code'] != 
		$transaction['code']) continue;
	    if ($account['transactions'][$key]['comment'] != 
		$transaction['comment']) continue;
	    if ($account['transactions'][$key]['amount'] != 
		$transaction['amount']) continue;
	    # echo "<!-- account_find_transaction: found $key -->\n";
	    return $key;
	}
    }
    # echo "<!-- account_find_transaction: nothing found -->\n";
    return -1;
}

# default sac file
function sac_default () {
    global $SERVER_NAME;

    # echo "<!-- sac_default: $SERVER_NAME -->\n";
    if ($SERVER_NAME == 'server') return "/var/log/wtmp";
    if ($SERVER_NAME == 'emma') return "/root/gic/wtmp-2000-11-partial";
    return "/var/log/wtmp";
}

# read transactions from system accounting records
function sac_read ($account, $file) {
    set_time_limit(30);
    # echo "\n";

    # generate per user clipped usage transactions
    $cm = './sac.`uname -m` -w '.$file.' -p -c -u '.$account['name'].' -l';
    # echo "<!-- sac_read: command: $cm -->\n";
    $pp = popen($cm, 'r');

    # decode first line, user total
    $line = chop(fgets($pp, 128));
    # echo "<!-- sac_read: header: $line -->\n";
    list($user, $hour) = split('  *', $line);

    $count = 0;

    # process remaining lines, transactions
    while (!feof($pp)) {
	$line = chop(fgets($pp, 128));
	# echo "<!-- sac_read: detail: $line -->\n";
	$count++;
	list($blank, $day, $month, $year, $start, $dash, $end, $port, $thing, $host, $hours) = split('  *', $line);
	if ($month == "") {
	    # echo "<!-- sac_read: skipped: month missing -->\n";
	    continue;
	}
	if ($hours == 0) {
	    # echo "<!-- sac_read: skipped: hours zero -->\n";
	    continue;
	}

	# convert the 'dd mmm yyyy hh:mm:ss' format to unix date time
	$month = date('m',strtotime($day.' '.$month.' '.$year));
	list($hour, $minute, $second) = explode(':', $end);
	$date = mktime($hour, $minute, $second, $month, $day, $year);
	$transaction['when'] = $date;
	$transaction['code'] = 'Computer';

        # ??? encode hours as hh:mm and h.hh
	# build a transaction description
	$transaction['comment'] = 'On '.$host.' from '.$start.' at '.amount_to_string($account['rate']).'/hr for '.$hours.' hours';
	$transaction['amount'] = amount_round_5c(- amount_from_string($hours) * amount_from_string($account['rate']) / 10000);
	$transaction['reference'] = $file.' '.sprintf("%06d", $count);

	# add the transaction if it isn't already on file
        if (account_find_transaction($account, $transaction) == -1) {
            $account['transactions'][] = $transaction;
	    # echo "<!-- sac_read: added -->\n";
	} else {
	    # echo "<!-- sac_read: duplicate -->\n";
	}

	# ??? possibility: someone is logged in at the time of the rollover
	# and is assigned one transaction now, and another later, need to
	# test what happens as a result; is the logout that is in the second
	# wtmp file properly reported?

    }

    pclose($pp);
    # echo "\n<!-- sac_read: count: ".$count." -->\n\n";
    return $account;
}

function compare_date_forward ($a, $b) {
    if ($a['when'] == $b['when']) {
	if (!isset($a['reference'])) return 0;
	if (!isset($b['reference'])) return 0;
	if ($a['reference'] == $b['reference']) return 0;
	return ($a['reference'] < $b['reference']);
    }
    return ($a['when'] < $b['when']) ? -1 : 1;
}

function compare_date_backward ($a, $b) {
    if ($a['when'] == $b['when']) {
	if (!isset($a['reference'])) return 0;
	if (!isset($b['reference'])) return 0;
	if ($a['reference'] == $b['reference']) return 0;
	return ($a['reference'] > $b['reference']);
    }
    return ($a['when'] > $b['when']) ? -1 : 1;
}

function account_total($name) {
    $total = 0;

    if (account_exists($name)) {
	$account = account_read($name);
    } else {
	$account = account_default($name);
    }

    $account = sac_read($account, sac_default());

    if (isset($account['transactions'])) {
	uasort($account['transactions'], compare_date_forward);
	reset ($account['transactions']);
	while (list ($key, $val) = each ($account['transactions'])) {
	    $total = $total + $account['transactions'][$key]['amount'];
	}
    }

    return $total;
}

# capture active users
function list_active_users(&$users) {
    $pp = popen('who', 'r');
    while (!feof($pp)) {
	list($user, $smb, $month, $day, $time, $host) = 
	    split('  *', chop(fgets($pp,128)));
	if ($user == "") continue;
	$users[$user]['name'] = $user;
    }
    pclose($pp);
}

# capture accounts defined
function list_accounts_defined(&$users) {
    $list = account_list();
    while (list ($key, $val) = each ($list)) {
	$users[$val]['name'] = $val;
    }
}

# capture from /etc/passwd
function list_known_users(&$users) {
    $fp = fopen('/etc/passwd', 'r');
    while (!feof($fp)) {
	$line = fgets($fp, 256);
	if ($line == "") continue;
	list($user, $pass, $uid, $gid, $gecos, $home) = explode(':', $line);
	list(, $root, $path) = explode('/', $home);
	if ($root != 'home') continue;
	if ($path == '') continue;
	list($name) = explode(',', $gecos);
	$users[$user]['name'] = $name;
    }
    fclose($fp);
}

function head($name) {
    echo TITLE;
    echo '<br>';
    echo '
<table border=0><tr>
<td valign=top>
<img src=gic-usage-accounting.jpg width=232 height=56 alt="Usage Accounting">
</td>
<td><img src=gic-slash.jpg border=0></td>
<td valign=top>
<img border=0 src='.$name.' alt='.$name.'>
</td>
</tr></table>
';
}

function icon($link, $icon) {
    global $SCRIPT_NAME;

    # derive a meaningful alternate name for the icon
    ereg('(.*)\.(.*)', $link, $regs);
    $name = ucfirst($regs[1]);

    if (eregi(basename($SCRIPT_NAME), $link)) {
	echo "<td valign=top align=center><a href=$link><img border=0 src=$icon alt=\"$name\"></a><br>^</td>";
    } else {
	echo "<td valign=top><a href=$link><img border=0 src=$icon alt=\"$name\"></a></td>";
    }
}

function slash() {
    echo '<td valign=top><img src=gic-slash.jpg border=0></td>';
}

function menu($name) {
    if (authenticated()) {
	echo '<center><table border=0><tr>';
	icon('menu.phtml', 'gic-menu.jpg');
	slash();
	icon('list.phtml', 'gic-list.jpg');
	if ($name != '') {
	    slash();
	    icon("view.phtml?name=$name", 'gic-view.jpg');
	    slash();
	    icon("payment.phtml?name=$name", 'gic-payment.jpg');
	    slash();
	    icon("charge.phtml?name=$name", 'gic-charge.jpg');
	    slash();
	    icon("modify.phtml?name=$name", 'gic-modify.jpg');
	}
	slash();
	icon('export.phtml', 'gic-export.jpg');
	slash();
	icon('codes.phtml', 'gic-codes.jpg');
	echo '</tr></table></center>';
    }
}

function codes_emit($code, $line, &$count) {
    list($item, $description) = split(' ', $line);
    echo '<option ';
    if ($code == $item) {
	echo 'selected ';
	$count++;
    }
    echo 'value="'.$item.'">'.$line.'</option>';
}

function codes_write($codes) {
    if (@is_file(PATH_CODES)) {
	if (!@is_writeable(PATH_CODES)) {
	    echo 'Codes file <tt>'.PATH_CODES.'</tt> is not writeable.';
	    return 0;
	}
    }
    if (!($fp = fopen(PATH_CODES, 'w'))) {
	echo 'Failed to open codes file.';
	return 0;
    }
    if (!flock($fp, 2)) {
	echo 'Failed to lock codes file.';
	fclose($fp);
	return 0;
    }
    if (!fwrite($fp,ereg_replace("\r\n", "\n", $codes))) {
	echo 'Failed to write to codes file.';
	fclose($fp);
	return 0;
    }
    if (!fclose($fp)) {
	echo 'Failed to close codes file.';
	return 0;
    }
    return 1;
}

function codes_read() {
    if (!@is_readable(PATH_CODES)) {
        $codes[] = 'CAR - Additional Swipe Card';
        $codes[] = 'CD - CD Burning';
        $codes[] = 'CON - Consumables';
        $codes[] = 'FAX - Facsimile';
        $codes[] = 'MEM - Membership';
        $codes[] = 'PHO - Photocopying';
        $codes[] = 'PRI - Printing';
        $codes[] = 'SCA - Scanning';
        $codes[] = 'ADJ - Adjustment';
        $codes[] = 'DIS - Discount';
        $codes[] = 'DON - Donations';
        $codes[] = 'MIS - Miscellaneous';
        $codes[] = 'SUP - Supervisors';
    } else {
	$fp = fopen(PATH_CODES, 'r');
	while ($line = chop(fgets($fp, 128))) {
	    if ($line != '') $codes[] = $line;
	}
	fclose($fp);
    }
    return $codes;
}

function codes_issue($code) {
    $count = 0;
    $codes = codes_read();
    reset($codes);
    echo '<select name=code size='.(count($codes)+1).'>';
    while (list(, $line) = each($codes)) {
	codes_emit($code, $line, $count);
    }
    if ($count == 0) {
	codes_emit($code, $code, $count);
    }
    echo '</select>';
}

function codes_form() {
    echo '<form method=post>';
    $codes = codes_read();
    echo '<textarea name=codes cols=40 rows='.(count($codes)+1).'>';
    reset($codes);
    while (list(, $line) = each($codes)) {
	echo $line."\n";
    }
    echo '</textarea><p>';
    echo '<input type=submit name=action value="Store Item Codes">';
    echo '</form>';
}

?>
